import math
from dataclasses import dataclass
from typing import Optional, Tuple, Union

import torch
import torch.utils.checkpoint
from torch import nn
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from transformers.activations import ACT2FN
from transformers.modeling_outputs import (
    BaseModelOutput,
    BaseModelOutputWithPooling,
    MaskedLMOutput,
    SequenceClassifierOutput,
)
from transformers.modeling_utils import PreTrainedModel
from transformers.pytorch_utils import apply_chunking_to_forward, find_pruneable_heads_and_indices, prune_linear_layer
from transformers.utils import (
    ModelOutput,
    add_code_sample_docstrings,
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
    logging,
)

from .configuration_bleurt import BleurtConfig


logger = logging.get_logger(__name__)

_CHECKPOINT_FOR_DOC = "lucadiliello/bleurt-base-128"
_CONFIG_FOR_DOC = "BleurtConfig"
_TOKENIZER_FOR_DOC = "BleurtTokenizer"


BERT_PRETRAINED_MODEL_ARCHIVE_LIST = [
    "lucadiliello/BLEURT-20",
    "lucadiliello/BLEURT-20-D12",
    "lucadiliello/BLEURT-20-D3",
    "lucadiliello/BLEURT-20-D6",
    "lucadiliello/bleurt-base-128",
    "lucadiliello/bleurt-base-512",
    "lucadiliello/bleurt-large-128",
    "lucadiliello/bleurt-large-512",
    "lucadiliello/bleurt-tiny-128",
    "lucadiliello/bleurt-tiny-512",
]


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L180  # noqa: E501
class BleurtEmbeddings(nn.Module):
    r""" Construct the embeddings from word, position and token_type embeddings. """

    def __init__(self, config: BleurtConfig):
        super().__init__()

        target_size = config.embedding_size if config.embedding_size is not None else config.hidden_size

        self.word_embeddings = nn.Embedding(config.vocab_size, target_size, padding_idx=config.pad_token_id)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, target_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, target_size)

        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = nn.LayerNorm(target_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        # position_ids (1, len position emb) is contiguous in memory and exported when serialized
        self.position_embedding_type = getattr(config, "position_embedding_type", "absolute")
        self.register_buffer("position_ids", torch.arange(config.max_position_embeddings).expand((1, -1)))
        self.register_buffer(
            "token_type_ids", torch.zeros(self.position_ids.size(), dtype=torch.long), persistent=False
        )

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        past_key_values_length: int = 0,
    ) -> torch.Tensor:

        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            input_shape = inputs_embeds.size()[:-1]

        seq_length = input_shape[1]

        if position_ids is None:
            position_ids = self.position_ids[:, past_key_values_length : seq_length + past_key_values_length]

        # Setting the token_type_ids to the registered buffer in constructor where it is all zeros, which usually
        # occurs when its auto-generated, registered buffer helps users when tracing the model without passing
        # token_type_ids, solves issue #5664
        if token_type_ids is None:
            if hasattr(self, "token_type_ids"):
                buffered_token_type_ids = self.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(input_shape[0], seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=self.position_ids.device)

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = inputs_embeds + token_type_embeddings

        position_embeddings = self.position_embeddings(position_ids)
        embeddings += position_embeddings

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L242  # noqa: E501
class BleurtSelfAttention(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()

        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                f"The hidden size ({config.hidden_size}) is not a multiple of the number of attention "
                f"heads ({config.num_attention_heads})"
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x: torch.Tensor) -> torch.Tensor:
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor]:

        query_layer = self.transpose_for_scores(self.query(hidden_states))
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)

        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
            attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(new_context_layer_shape)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        return outputs


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L376  # noqa: E501
class BleurtSelfOutput(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: torch.Tensor, input_tensor: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


# moslty copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L390  # noqa: E501
class BleurtAttention(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.self = BleurtSelfAttention(config)
        self.output = BleurtSelfOutput(config)
        self.pruned_heads = set()

    def prune_heads(self, heads):
        if len(heads) == 0:
            return
        heads, index = find_pruneable_heads_and_indices(
            heads, self.self.num_attention_heads, self.self.attention_head_size, self.pruned_heads
        )

        # Prune linear layers
        self.self.query = prune_linear_layer(self.self.query, index)
        self.self.key = prune_linear_layer(self.self.key, index)
        self.self.value = prune_linear_layer(self.self.value, index)
        self.output.dense = prune_linear_layer(self.output.dense, index, dim=1)

        # Update hyper params and store pruned heads
        self.self.num_attention_heads = self.self.num_attention_heads - len(heads)
        self.self.all_head_size = self.self.attention_head_size * self.self.num_attention_heads
        self.pruned_heads = self.pruned_heads.union(heads)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor]:

        self_outputs = self.self(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            head_mask=head_mask,
            output_attentions=output_attentions,
        )

        attention_output = self.output(self_outputs[0], hidden_states)
        outputs = (attention_output,) + self_outputs[1:]  # add attentions if we output them
        return outputs


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L439  # noqa: E501
class BleurtIntermediate(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)

        if isinstance(config.hidden_act, str):
            self.intermediate_act_fn = ACT2FN[config.hidden_act]
        else:
            self.intermediate_act_fn = config.hidden_act

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L454  # noqa: E501
class BleurtOutput(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: torch.Tensor, input_tensor: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L468  # noqa: E501
class BleurtLayer(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()

        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = BleurtAttention(config)
        self.intermediate = BleurtIntermediate(config)
        self.output = BleurtOutput(config)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor]:

        self_attention_outputs = self.attention(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            head_mask=head_mask,
            output_attentions=output_attentions,
        )

        attention_output = self_attention_outputs[0]
        outputs = self_attention_outputs[1:]  # add self attentions if we output attention weights

        layer_output = apply_chunking_to_forward(
            self.feed_forward_chunk, self.chunk_size_feed_forward, self.seq_len_dim, attention_output
        )

        outputs = (layer_output,) + outputs
        return outputs

    def feed_forward_chunk(self, attention_output):
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L554  # noqa: E501
class BleurtEncoder(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList([BleurtLayer(config) for _ in range(config.num_hidden_layers)])
        self.gradient_checkpointing = False

        if config.embedding_size is not None:
            self.embedding_projection = nn.Linear(config.embedding_size, config.hidden_size)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
        output_hidden_states: Optional[bool] = False,
    ) -> BaseModelOutput:

        if hasattr(self, "embedding_projection"):
            hidden_states = self.embedding_projection(hidden_states)

        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None

        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None

            if self.gradient_checkpointing and self.training:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs, output_attentions)

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    output_attentions,
                )

            hidden_states = layer_outputs[0]
            if output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        return BaseModelOutput(
            last_hidden_state=hidden_states,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L651  # noqa: E501
class BleurtPooler(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L666  # noqa: E501
class BleurtPredictionHeadTransform(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()

        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        if isinstance(config.hidden_act, str):
            self.transform_act_fn = ACT2FN[config.hidden_act]
        else:
            self.transform_act_fn = config.hidden_act
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L683  # noqa: E501
class BleurtLMPredictionHead(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()

        self.transform = BleurtPredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        self.bias = nn.Parameter(torch.zeros(config.vocab_size))

        # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`
        self.decoder.bias = self.bias

    def forward(self, hidden_states):
        hidden_states = self.transform(hidden_states)
        hidden_states = self.decoder(hidden_states)
        return hidden_states


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L703  # noqa: E501
class BleurtOnlyMLMHead(nn.Module):

    def __init__(self, config: BleurtConfig):
        super().__init__()
        self.predictions = BleurtLMPredictionHead(config)

    def forward(self, sequence_output: torch.Tensor) -> torch.Tensor:
        prediction_scores = self.predictions(sequence_output)
        return prediction_scores


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L735  # noqa: E501
class BleurtPreTrainedModel(PreTrainedModel):
    r""" An abstract class to handle weights initialization and a simple interface for downloading and
    loading pretrained models. """

    config_class = BleurtConfig
    base_model_prefix = "bleurt"
    supports_gradient_checkpointing = True
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def _init_weights(self, module):
        r""" Initialize the weights. """

        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()

        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()

        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def _set_gradient_checkpointing(self, module, value=False):
        if isinstance(module, BleurtEncoder):
            module.gradient_checkpointing = value


@dataclass
class BleurtForPreTrainingOutput(ModelOutput):
    r"""
    Output type of [`BleurtForPreTraining`].

    Args:
        loss (*optional*, returned when `labels` is provided, `torch.FloatTensor` of shape `(1,)`):
            Total loss as the sum of the masked language modeling loss and the next sequence prediction
            (classification) loss.
        prediction_logits (`torch.FloatTensor` of shape `(batch_size, sequence_length, config.vocab_size)`):
            Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        hidden_states (`tuple(torch.FloatTensor)`, *optional*, returned when `output_hidden_states=True` is passed or
            when `config.output_hidden_states=True`):
            Tuple of `torch.FloatTensor` (one for the output of the embeddings + one for the output of each layer) of
            shape `(batch_size, sequence_length, hidden_size)`.

            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        attentions (`tuple(torch.FloatTensor)`, *optional*, returned when `output_attentions=True` is passed or when
            `config.output_attentions=True`):
            Tuple of `torch.FloatTensor` (one for each layer) of shape `(batch_size, num_heads, sequence_length,
            sequence_length)`.

            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
    """

    loss: Optional[torch.FloatTensor] = None
    prediction_logits: torch.FloatTensor = None
    hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[torch.FloatTensor]] = None


BLEURT_START_DOCSTRING = r"""

    This model inherits from [`PreTrainedModel`]. Check the superclass documentation for the generic methods the
    library implements for all its model (such as downloading or saving, resizing the input embeddings, pruning heads
    etc.)

    This model is also a PyTorch [torch.nn.Module](https://pytorch.org/docs/stable/nn.html#torch.nn.Module) subclass.
    Use it as a regular PyTorch Module and refer to the PyTorch documentation for all matter related to general usage
    and behavior.

    Parameters:
        config ([`BleurtConfig`]): Model configuration class with all the parameters of the model.
            Initializing with a config file does not load the weights associated with the model, only the
            configuration. Check out the [`~PreTrainedModel.from_pretrained`] method to load the model weights.
"""

BLEURT_INPUTS_DOCSTRING = r"""
    Args:
        input_ids (`torch.LongTensor` of shape `({0})`):
            Indices of input sequence tokens in the vocabulary.

            Indices can be obtained using [`BertTokenizer`]. See [`PreTrainedTokenizer.encode`] and
            [`PreTrainedTokenizer.__call__`] for details.

        attention_mask (`torch.FloatTensor` of shape `({0})`, *optional*):
            Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.

        token_type_ids (`torch.LongTensor` of shape `({0})`, *optional*):
            Segment token indices to indicate first and second portions of the inputs. Indices are selected in `[0,
            1]`:

            - 0 corresponds to a *sentence A* token,
            - 1 corresponds to a *sentence B* token.

        position_ids (`torch.LongTensor` of shape `({0})`, *optional*):
            Indices of positions of each input sequence tokens in the position embeddings. Selected in the range `[0,
            config.max_position_embeddings - 1]`.

        head_mask (`torch.FloatTensor` of shape `(num_heads,)` or `(num_layers, num_heads)`, *optional*):
            Mask to nullify selected heads of the self-attention modules. Mask values selected in `[0, 1]`:

            - 1 indicates the head is **not masked**,
            - 0 indicates the head is **masked**.

        inputs_embeds (`torch.FloatTensor` of shape `({0}, hidden_size)`, *optional*):
            Optionally, instead of passing `input_ids` you can choose to directly pass an embedded representation. This
            is useful if you want more control over how to convert `input_ids` indices into associated vectors than the
            model's internal embedding lookup matrix.
        output_attentions (`bool`, *optional*):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under returned
            tensors for more detail.
        output_hidden_states (`bool`, *optional*):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors for
            more detail.
"""


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L872  # noqa: E501
@add_start_docstrings(
    "The bare Bert Model transformer outputting raw hidden-states without any specific head on top.",
    BLEURT_START_DOCSTRING,
)
class BleurtModel(BleurtPreTrainedModel):
    r""" The model can behave only as an encoder. """

    def __init__(self, config: BleurtConfig, add_pooling_layer: bool = True):
        super().__init__(config)
        self.config = config

        self.embeddings = BleurtEmbeddings(config)
        self.encoder = BleurtEncoder(config)

        self.pooler = BleurtPooler(config) if add_pooling_layer else None

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    @add_start_docstrings_to_model_forward(BLEURT_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        processor_class=_TOKENIZER_FOR_DOC,
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=BaseModelOutputWithPooling,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> BaseModelOutputWithPooling:

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        batch_size, seq_length = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length)), device=device)

        if token_type_ids is None:
            if hasattr(self.embeddings, "token_type_ids"):
                buffered_token_type_ids = self.embeddings.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(batch_size, seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape)

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
        )

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        sequence_output = encoder_outputs.last_hidden_state
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        return BaseModelOutputWithPooling(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L1301  # noqa: E501
@add_start_docstrings("""Bert Model with a `language modeling` head on top.""", BLEURT_START_DOCSTRING)
class BleurtForMaskedLM(BleurtPreTrainedModel):

    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids", r"predictions.decoder.bias", r"cls.predictions.decoder.weight"]

    def __init__(self, config: BleurtConfig):
        super().__init__(config)

        self.bleurt = BleurtModel(config, add_pooling_layer=False)
        self.cls = BleurtOnlyMLMHead(config)

        # Initialize weights and apply final processing
        self.post_init()

    def get_output_embeddings(self):
        return self.cls.predictions.decoder

    def set_output_embeddings(self, new_embeddings):
        self.cls.predictions.decoder = new_embeddings

    @add_start_docstrings_to_model_forward(BLEURT_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        processor_class=_TOKENIZER_FOR_DOC,
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=MaskedLMOutput,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], MaskedLMOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for computing the masked language modeling loss. Indices should be in `[-100, 0, ...,
            config.vocab_size]` (see `input_ids` docstring) Tokens with indices set to `-100` are ignored (masked), the
            loss is only computed for the tokens with labels in `[0, ..., config.vocab_size]`
        """

        outputs = self.bleurt(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        sequence_output = outputs.last_hidden_state
        prediction_scores = self.cls(sequence_output)

        masked_lm_loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()  # -100 index = padding token
            masked_lm_loss = loss_fct(prediction_scores.view(-1, self.config.vocab_size), labels.view(-1))

        return MaskedLMOutput(
            loss=masked_lm_loss,
            logits=prediction_scores,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


# mostly copied from https://github.com/huggingface/transformers/blob/00ba7cadd812437708b380ab078a3cfe8cfaff31/src/transformers/models/bert/modeling_bert.py#L1518  # noqa: E501
@add_start_docstrings(
    """
    Bert Model transformer with a sequence classification/regression head on top (a linear layer on top of the pooled
    output) e.g. for GLUE tasks.
    """,
    BLEURT_START_DOCSTRING,
)
class BleurtForSequenceClassification(BleurtPreTrainedModel):

    def __init__(self, config: BleurtConfig):
        super().__init__(config)

        self.num_labels = config.num_labels
        self.config = config

        self.bleurt = BleurtModel(config)

        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )

        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    @add_start_docstrings_to_model_forward(BLEURT_INPUTS_DOCSTRING.format("batch_size, sequence_length"))
    @add_code_sample_docstrings(
        processor_class=_TOKENIZER_FOR_DOC,
        output_type=SequenceClassifierOutput,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], SequenceClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """

        outputs = self.bleurt(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        pooled_output = self.dropout(outputs.pooler_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



import traceback

from vero.metrics import MetricBase
import numpy as np
import math
# from rag_code.tracing_components import logger

def true_ranks(reranked_docs: list, ground_truth: list) -> list | None:
    # logger.info('Calculating ranks...')
    try:
        ranks = []
        for re_doc, truth in zip(reranked_docs, ground_truth):
            s = []
            for doc in re_doc:
                if doc in truth:
                    s.append(truth[doc])
                else:
                    s.append(0)
            ranks.append(s)
        return ranks
    except Exception as e:
        print('Exception occured during rank calculation\nError:', traceback.format_exc())
        # logger.info('Exception occured during rank calculation\nError:', e)
        return None


class CumulativeNDCG(MetricBase):
    '''
        Unique implementation of NDCG that can be used to evaluate the cumulative performance of retriever and reranker.

        :param reranked_docs: Pass the list of reranked documents list.
        :param ground_truth: Pass the ground truth list for comparison.

        Methods
        ---------
        1. __init__(reranked_docs, ground_truth)
            Initializes the metric.
        2. evaluate() -> list
            Returns the ndcg score.

        :returns: list
        '''
    name = 'cumulative_ndcg'

    def __init__(self,reranked_docs: list, ground_truth: list):
        self.reranked_docs = reranked_docs
        self.ground_truth = ground_truth

    def evaluate(self) -> list | None:
        # logger.info(f'Starting ndcg calculation...')
        ranks = true_ranks(self.reranked_docs, self.ground_truth)
        ndcg = []
        try:
            for rank, truth in zip(ranks, self.ground_truth):
                dcg = 0
                idcg = 0
                k = len(rank)
                for i in range(k):
                    dcg += (2 ** (rank[i]) - 1) / math.log2(2 + i)

                truth_ranks = sorted(truth.values(), reverse=True)

                limit = min(k, len(truth_ranks))
                for i in range(limit):
                    idcg += (2 ** (truth_ranks[i]) - 1) / math.log2(2 + i)
                if idcg == 0:
                    ndcg.append(0)
                else:
                    ndcg.append(round(dcg / idcg, 2))
            return ndcg
        except Exception as e:
            print('Exception occured during cumulative NDCG calculation\nError:', traceback.format_exc())
            # logger.info('Exception occured during NDCG calculation\nError:', e)
            return None


import os
from typing import Union

from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from .tokenization_bleurt_fast import BleurtTokenizerFast
from .tokenization_bleurt_sp import BleurtSPTokenizer


class BleurtTokenizer(PreTrainedTokenizerBase):
    r"""
    Construct a BLEURT tokenizer and returns either `BleurtTokenizerFast` or `BleurtSPTokenizer` based
    on whether the model requires sentencepiece tokenization or not.
    """

    def __new__(cls, *args, **kwargs) -> Union[BleurtTokenizerFast, BleurtSPTokenizer]:
        try:
            return BleurtTokenizerFast(*args, **kwargs)
        except (OSError, TypeError):
            return BleurtSPTokenizer(*args, **kwargs)

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], *init_inputs, **kwargs
    ) -> Union[BleurtTokenizerFast, BleurtSPTokenizer]:
        try:
            return BleurtTokenizerFast.from_pretrained(pretrained_model_name_or_path, *init_inputs, **kwargs)
        except (OSError, TypeError):
            return BleurtSPTokenizer.from_pretrained(pretrained_model_name_or_path, *init_inputs, **kwargs)



report_review_prompt='''

ROLE:
You are a meticulous Quality Assurance (QA) Analyst and Fact-Checker. You are tasked with the final review of a technical report. Your standards are exceptionally high, and your sole purpose is to ensure that the final report is a perfectly faithful, accurate, and logical representation of the source data. You are skeptical by nature and trust nothing in the report without verifying it against the source data.

GOAL:
Your goal is to review the final generated Markdown report against the collection of "source-of-truth" JSON data from the preceding analytical agents. You must identify any discrepancies, including factual errors, logical fallacies, omissions, or hallucinations. Based on your findings, you will produce a structured JSON feedback object and flag whether the report needs to be regenerated.

INPUT DESCRIPTION:
You will receive two sets of inputs:

Source_of_Truth_JSONs: A collection of all JSON objects generated by Agents 1 through 4. This includes the EvaluationContext, the individual ComponentAnalysis reports, the HolisticDiagnosis, and the Recommendations. This is the ground truth.

Report_Under_Review: The final, human-readable report in Markdown format generated by Agent 5. This is the document you must critique.

INPUTS:
Source of Truth JSONs: {data}
Report Under Review: {generated_report}

VERIFICATION CHECKLIST & HEURISTICS:
You must systematically review the report by following this checklist.

1. Data Fidelity Check (Highest Priority):

Action: For every numerical value, metric name, score, or specific model name mentioned in the report, find its corresponding value in the Source_of_Truth_JSONs and verify that they match exactly.

Look For: Mismatched numbers (e.g., report says Recall@5 is 0.82, but JSON says 0.88), typos in metric names, incorrect model names. These are Data Errors.

2. Logical Consistency Check:

Action: Read the conclusions and narratives in the report (especially the Executive Summary and Holistic Diagnosis sections). Verify that these conclusions are logically supported by the HolisticDiagnosis and ComponentAnalysis JSONs.

Look For: The report blaming one component when the diagnosis blamed another; the summary highlighting a strength that was diagnosed as a weakness. These are Logical Errors.

3. Completeness Check (Omissions):

Action: Compare the key findings in the HolisticDiagnosis and Recommendations JSONs against the final report.

Look For: Missing primary_weaknesses, missing causal_chains, or missing high-priority improvement_suggestions. These are Omissions.

4. Causality Check:

Action: Pay special attention to the "Causal Chains" section of the report. Verify that the cause-and-effect relationships described are an accurate reflection of the causal_chains array in the HolisticDiagnosis JSON.

Look For: Misrepresentation of the root cause of a pipeline failure. This is a severe Logical Error.

LOGIC FOR THE regeneration_needed FLAG:
You must set the output flag based on the severity of the issues found:

Set regeneration_needed: true IF: You find one or more "Data Error," "Logical Error," or significant "Omission." The report is factually incorrect or misleading and MUST be fixed.

Set regeneration_needed: false IF: The report is perfect, OR if you only find minor stylistic or formatting issues that do not affect the meaning or accuracy of the report.

OUTPUT SPECIFICATION:
Your final output MUST be a single, valid JSON object and nothing else. It must conform to this exact structure:
{{
  "regeneration_needed": true,
  "overall_assessment": "A one-two sentence summary of the report's quality.",
  "feedback_points": [
    {{
      "type": "Data Error | Logical Error | Omission | Formatting",
      "location_quote": "A direct quote from the report where the issue was found.",
      "issue_description": "A clear and concise explanation of what is wrong.",
      "evidence_from_source": "The specific value or finding from the source JSON that contradicts the report.",
      "suggested_correction": "A clear instruction on how to fix the issue."
    }}
  ]
}}

SELF-REFLECTION & CONFIDENCE ASSESSMENT:
After completing your verification checklist and before generating your final feedback JSON, perform this meta-cognitive step:
- Feedback Assessment: "Based on the feedback I have given, is feedback actually useful and actionable, are there any gaps in my feedback or is there any wrong feedback hidden in there?"
- Confidence Score: "On a scale of 1 to 10 (where 10 is perfect), what is my confidence score that this report is accurate and ready for a human reader without changes?"
- Flag Justification: "Does my regeneration_needed flag (true/false) accurately reflect this confidence score? Have I followed the logic rules for setting this flag without exception?"

'''


from .bleurt.configuration_bleurt import BleurtConfig  # noqa: F401
from .bleurt.modeling_bleurt import BleurtForSequenceClassification  # noqa: F401
from .bleurt.tokenization_bleurt import BleurtTokenizer  # noqa: F401



import pandas as pd
import numpy as np
import os
import re
from openai import OpenAI
from langchain_community.document_loaders import DirectoryLoader, PyPDFLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter
import random
import json
from dotenv import load_dotenv, find_dotenv
import langgraph
from datetime import date
from fastapi import FastAPI
from pydantic import BaseModel
import time
from typing import TypedDict, List, Dict, Any, Optional
from langgraph.graph import StateGraph, START, END
import sys
import requests
from pprint import pprint
import datetime
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger
import asyncio
from typing import List, Literal
from pydantic import BaseModel, Field, conlist
from tqdm import tqdm

pd.set_option('display.max_colwidth', 200)
load_dotenv(find_dotenv())

client = OpenAI()

from vero.test_dataset_generator import chunking_utilities

from vero.test_dataset_generator.prompts import system_prompt_basic, system_prompt_chunk_boundary, system_prompt_chunk_length, \
    system_prompt_query_intent, user_prompt_chunks, system_prompt_persona_generation, \
        user_prompt_personas, system_prompt_persona_specific_QA


def get_openai_resp_struct(system_prompt: str, user_prompt: str, info_inp: dict, \
                           resp_format, model_id: str = "o3-mini-2025-01-31"):
    """
    Build and call the OpenAI Responses API, returning a Pydantic-validated structure.

    Purpose:
    - Format the provided user prompt with `info_chunk_inp`, call the responses.parse
      endpoint and return the parsed output (preferably a Pydantic model instance).
    Inputs:
    - system_prompt: The system-level instruction string to bias behavior.
    - user_prompt: The user-level prompt template (must include '{info_chunk}' when used that way).
    - info_chunk_inp: The chunk payload (already JSON-serializable or a string) to inject.
    - resp_format: A Pydantic model class or structure that the SDK will use to validate the response.
    - model_id: Model identifier string used for the Responses API call.
    Returns:
    - The parsed output (usually a Pydantic model instance) if parsing succeeded, otherwise the raw response.
    Notes:
    - The function relies on the `client` OpenAI instance already created.
    """
    if info_inp and user_prompt:
        formatted_user = user_prompt.format(inp=info_inp)
    else:
        formatted_user = "Start the task"
    response = client.responses.parse(
        model=model_id,
        input=formatted_user,  # user prompt (runtime-filled)
        instructions=(
                system_prompt
                + "\n\n[STRUCTURE] Respond ONLY as JSON matching the provided schema. "
        ),
        text_format=resp_format,  # Pydantic model (schema)
        max_output_tokens=50000
    )
    return getattr(response, "output_parsed", response)


class QAItem_basic(BaseModel):
    """
    Schema for a single basic QA item.

    Fields:
    - question: The question text. Must NOT mention chunk IDs.
    - answer: A single, unambiguous factual answer supported by the passages.
    - chunk_ids: List of relevant chunk IDs that support the answer.
    - difficulty: Declared difficulty level: "Easy", "Medium", or "Hard".
    """
    question: str = Field(..., description="The question text. Must NOT mention chunk IDs.")
    answer: str = Field(..., description="A single, unambiguous factual answer supported by the passages.")
    chunk_ids: List[str] = Field(
        ..., description="List of relevant chunk IDs (strings) that support the answer."
    )
    difficulty: Literal["Easy", "Medium", "Hard"] = Field(
        ..., description="Declared difficulty level."
    )

class QAResponse_basic(BaseModel):
    """
    Response container for a group of basic QA items.

    Constraints:
    - Contains exactly 5 items (1 Easy, 2 Medium, 2 Hard) as enforced by the calling prompts.
    """
    # Exactly 5 items (1 Easy, 2 Medium, 2 Hard). The model is told to respect this in the instructions.
    items: conlist(QAItem_basic, min_length=5, max_length=5)


def qaresponse_to_df_basic(qa_response):
    """
    Convert a QAResponse (Pydantic object) into a pandas DataFrame.

    Each row corresponds to a single QAItem_basic. This helper flattens the
    chunk ID lists into comma-separated strings for CSV-friendly output.
    """
    records = [
        {
            "Question": item.question,
            "Answer": item.answer,
            "Chunk IDs": ", ".join(item.chunk_ids),  # flatten list into string
            "Difficulty": item.difficulty
        }
        for item in qa_response.items
    ]
    return pd.DataFrame(records)


def get_QA_basic(dfct, n=10):
    """
    Generate basic QA items by sampling chunks and calling the model.

    Workflow:
    - Repeats calls to the Responses API until `n` items are gathered (requests are batched,
      each response is expected to contain exactly 5 items).
    - Samples a small set of chunks per call to provide grounding contexts.
    Inputs:
    - dfct: DataFrame of chunks with at least 'chunk_id' and 'text' columns.
    - n: Desired number of questions to produce.
    Returns:
    - DataFrame with generated QA items (columns: Question, Answer, Chunk IDs, Difficulty).
    """
    df_r = pd.DataFrame(columns=['Question', 'Answer', 'Chunk IDs', 'Difficulty'])
    for i1 in range(int(np.ceil(n/5))):
        sample_chunks = dfct.sample(20)
        cd = sample_chunks[['chunk_id', 'text']].to_dict(orient='records')
        resp1 = get_openai_resp_struct(system_prompt_basic, user_prompt_chunks, json.dumps(cd), QAResponse_basic)
        df1 = qaresponse_to_df_basic(resp1)
        df_r = pd.concat([df_r, df1])
        time.sleep(1)
    return df_r

class QAItem_len_bias(BaseModel):
    """
    Schema for QA items focusing on chunk-length bias.

    Fields:
    - question: The question text. Must NOT mention chunk IDs.
    - answer: A single, unambiguous factual answer supported by the passages.
    - more_relevant_chunk_ids: Chunks judged more relevant for the QA pair.
    - less_relevant_chunk_ids: Chunks judged less relevant.
    - short_rationale: Short reasoning that explains the discriminator favoring shorter chunks.
    - difficulty: Declared difficulty level.
    """
    question: str = Field(..., description="The question text. Must NOT mention chunk IDs.")
    answer: str = Field(..., description="A single, unambiguous factual answer supported by the passages.")
    more_relevant_chunk_ids: List[str] = Field(
        ..., description="List of more relevant chunk IDs (strings) that support the answer."
    )
    less_relevant_chunk_ids: List[str] = Field(
        ..., description="List of less relevant chunk IDs (strings) that support the answer."
    )
    short_rationale: str = Field(..., description="short reasoning highlighting discriminator favouring shorter chunks")
    difficulty: Literal["Easy", "Medium", "Hard"] = Field(
        ..., description="Declared difficulty level."
    )


class QAResponse_len_bias(BaseModel):
    """
    Response container for QA items that test chunk-length bias.

    Constraints:
    - Contains a fixed number of items (here min_length=8, max_length=8 as used in prompts).
    """
    # Exactly 5 items (1 Easy, 2 Medium, 2 Hard). The model is told to respect this in the instructions.
    items: conlist(QAItem_len_bias, min_length=8, max_length=8)


def qaresponse_to_df_len_bias(qa_response):
    """
    Convert a QAResponse_len_bias object into a DataFrame.

    Flattens both more/less relevant chunk id lists and includes the rationale field.
    """
    rows = []
    for item in getattr(qa_response, "items", []):
        rows.append({
            "Question": item.question,
            "Answer": item.answer,
            "Chunk IDs": ", ".join(map(str, getattr(item, "more_relevant_chunk_ids", []))),
            "Less Relevant Chunk IDs": ", ".join(map(str, getattr(item, "less_relevant_chunk_ids", []))),
            "Difficulty": item.difficulty,
            "Rationale": getattr(item, "short_rationale", None),
        })
    return pd.DataFrame(
        rows,
        columns=["Question", "Answer", "Chunk IDs",
                 "Less Relevant Chunk IDs", "Difficulty", "Rationale"]
    )



def get_QA_chunk_length(dfct, n=10):
    """
    Generate QA items that test sensitivity to chunk length.

    Workflow:
    - Scans cluster groups looking for clusters with multiple short chunks and
      satisfies other heuristics, then calls the model to produce QA items.
    Inputs:
    - dfct: DataFrame containing chunk metadata including 'cluster_id' and 'token_len'.
    - n: Number of questions requested (function batches calls tuned to prompt sizes).
    Returns:
    - DataFrame of generated QA items with length-bias annotations.
    """
    dfr = pd.DataFrame(columns=["Question", "Answer", "Chunk IDs", "Less Relevant Chunk IDs", "Difficulty", "Rationale"])
    for i1 in range(int(np.ceil(n/10))):
        lst = []
        
        for i in dfct['cluster_id'].unique():    ## find cluster of chunks with atleast 2 small chunks, long_chunks>=small_chunks and has not already been considered
            dft = dfct[dfct['cluster_id']==i]
            avg_len = np.average(dft['token_len'])
            dft1 = dft[dft['token_len']<50]
            if len(dft1)>1 and len(dft1)/len(dft)<0.5 and i not in lst:
                lst.append(i)
                break

        cd = dft[['chunk_id', 'text']].to_dict(orient='records')

        resp1 = get_openai_resp_struct(system_prompt_chunk_length, user_prompt_chunks, json.dumps(cd), QAResponse_len_bias)
        df1 = qaresponse_to_df_len_bias(resp1)

        dfr = pd.concat([dfr, df1])
        time.sleep(1)
        if len(lst)==dfct['cluster_id'].nunique():
            return dfr
    return dfr

class QAItem_boundary(BaseModel):
    """
    Schema for QA items that probe chunk boundary/synthesis behavior.

    Fields:
    - question: The question text. Must NOT mention chunk IDs.
    - answer: A single factual answer supported by passages.
    - chunk_ids: Relevant chunk identifiers.
    - difficulty: Declared difficulty level.
    - rationale: 1-2 sentences describing why this question tests boundary or synthesis behavior.
    """
    question: str = Field(..., description="The question text. Must NOT mention chunk IDs.")
    answer: str = Field(..., description="A single, unambiguous factual answer supported by the passages.")
    chunk_ids: List[str] = Field(
        ..., description="List of relevant chunk IDs (strings) that support the answer."
    )
    difficulty: Literal["Easy", "Medium", "Hard"] = Field(
        ..., description="Declared difficulty level."
    )
    rationale: str = Field(..., description="1-2 sentences about why this questions tests bounday/synthesis")


class QAResponse_boundary(BaseModel):
    """
    Container for boundary-testing QA items.

    Constraints:
    - Expected number of items as enforced by the calling prompt (here min_length=10, max_length=10).
    """
    # Exactly 5 items (1 Easy, 2 Medium, 2 Hard). The model is told to respect this in the instructions.
    items: conlist(QAItem_boundary, min_length=10, max_length=10)



def qaresponse_to_df_boundary(qa_response):
    """Each row = one QAItem_boundary. Joins chunk_ids; includes rationale."""
    rows = []
    for item in getattr(qa_response, "items", []):
        rows.append({
            "Question": item.question,
            "Answer": item.answer,
            "Chunk IDs": ", ".join(map(str, getattr(item, "chunk_ids", []))),
            "Difficulty": item.difficulty,
            "Rationale": getattr(item, "rationale", None),
        })
    return pd.DataFrame(rows, columns=["Question", "Answer", "Chunk IDs", "Difficulty", "Rationale"])


def get_QA_chunk_boundary(dfct, n=10):
    """
    Generate QA items that stress chunk boundary and synthesis handling.

    Workflow:
    - Looks for clusters with a mix of short/long chunks and requests boundary-testing items
      from the model using the boundary-focused system prompt.
    Inputs:
    - dfct: DataFrame of chunks with clustering information.
    - n: Number of questions requested.
    Returns:
    - DataFrame with boundary-focused QA items.
    """
    dfr = pd.DataFrame(columns=["Question","Answer","Chunk IDs","Difficulty","Rationale"])
    for i1 in range(int(np.ceil(n/10))):
        lst = []
        
        for i in dfct['cluster_id'].unique():    ## find cluster of chunks with atleast 2 small chunks, long_chunks>=small_chunks and has not already been considered
            dft = dfct[dfct['cluster_id']==i]
            avg_len = np.average(dft['token_len'])
            dft1 = dft[dft['token_len']<50]
            if len(dft1)>1 and len(dft1)/len(dft)<0.5 and i not in lst:
                lst.append(i)
                break

        cd = dft[['chunk_id', 'text']].to_dict(orient='records')
        # for i in range(len(dft)):
        #     cd[i] = dft.iloc[i]['text']

        resp1 = get_openai_resp_struct(system_prompt_chunk_boundary, user_prompt_chunks, json.dumps(cd), QAResponse_boundary)
        df1 = qaresponse_to_df_boundary(resp1)

        dfr = pd.concat([dfr, df1])
        time.sleep(1)
        if len(lst)==dfct['cluster_id'].nunique():
            return dfr
    return dfr


class QAItem_intent(BaseModel):
    """
    Schema for QA items that examine query intent understanding.

    Fields:
    - question: The question text. Must NOT mention chunk IDs.
    - answer: A single factual answer supported by the passages.
    - chunk_ids: Relevant chunk identifiers.
    - difficulty: Declared difficulty level.
    - rationale: 1-2 sentences about why the question tests intent understanding.
    """
    question: str = Field(..., description="The question text. Must NOT mention chunk IDs.")
    answer: str = Field(..., description="A single, unambiguous factual answer supported by the passages.")
    chunk_ids: List[str] = Field(
        ..., description="List of relevant chunk IDs (strings) that support the answer."
    )
    difficulty: Literal["Easy", "Medium", "Hard"] = Field(
        ..., description="Declared difficulty level."
    )
    rationale: str = Field(..., description="1-2 sentences about why this questions tests bounday/synthesis")


class QAResponse_intent(BaseModel):
    """
    Container for intent-focused QA items.

    Constraints:
    - Expected set size as defined by the prompt (here min_length=10, max_length=10).
    """
    # Exactly 5 items (1 Easy, 2 Medium, 2 Hard). The model is told to respect this in the instructions.
    items: conlist(QAItem_intent, min_length=10, max_length=10)



def qaresponse_to_df_intent(qa_response):
    """Each row = one QAItem_intent. Joins chunk_ids; includes rationale."""
    rows = []
    for item in getattr(qa_response, "items", []):
        rows.append({
            "Question": item.question,
            "Answer": item.answer,
            "Chunk IDs": ", ".join(map(str, getattr(item, "chunk_ids", []))),
            "Difficulty": item.difficulty,
            "Rationale": getattr(item, "rationale", None),
        })
    return pd.DataFrame(rows, columns=["Question", "Answer", "Chunk IDs", "Difficulty", "Rationale"])


def get_QA_query_intent(dfct, n=10):
    """
    Generate QA items designed to test query intent coverage.

    Workflow:
    - Selects cluster candidates similar to other generation helpers and asks the model to
      generate items that reveal whether intent is preserved or lost across chunking.
    Inputs:
    - dfct: DataFrame of chunks with clustering and token length metadata.
    - n: Number of questions requested.
    Returns:
    - DataFrame with intent-focused QA items.
    """
    dfr = pd.DataFrame(columns=["Question","Answer","Chunk IDs","Difficulty","Rationale"])
    for i1 in range(int(np.ceil(n/10))):
        lst = []
        
        for i in dfct['cluster_id'].unique():    ## find cluster of chunks with atleast 2 small chunks, long_chunks>=small_chunks and has not already been considered
            dft = dfct[dfct['cluster_id']==i]
            avg_len = np.average(dft['token_len'])
            dft1 = dft[dft['token_len']<50]
            if len(dft1)>1 and len(dft1)/len(dft)<0.5 and i not in lst:
                lst.append(i)
                break

        cd = dft[['chunk_id', 'text']].to_dict(orient='records')

        resp1 = get_openai_resp_struct(system_prompt_query_intent, user_prompt_chunks, json.dumps(cd), QAResponse_intent)
        df1 = qaresponse_to_df_intent(resp1)

        dfr = pd.concat([dfr, df1])
        time.sleep(1)
        if len(lst)==dfct['cluster_id'].nunique():
            return dfr
    return dfr

class TechSavviness(BaseModel):
    """
    Schema defining a persona's comfort and proficiency with technology.

    Fields:
    - level: Declared technology familiarity level — must be one of: 'Low', 'Medium', or 'High'.
    - justification: A concise, 1-sentence explanation describing why the persona fits the chosen tech level.
    """
    level: Literal["Low", "Medium", "High"] = Field(..., description="One of: 'Low', 'Medium', 'High'.")
    justification: str = Field(..., description="1-sentence explanation for their tech savviness level.")

class LanguageProficiency(BaseModel):
    """
    Schema describing a persona’s fluency and communication style in English.

    Fields:
    - level: Language proficiency category — must be one of: 'Native', 'Proficient', 'Conversational', or 'Limited'.
    - characteristic: A 1-sentence description of how this proficiency manifests in communication or writing style.
    """
    level: Literal["Native", "Proficient", "Conversational", "Limited"] = Field(..., description="Language level.")
    characteristic: str = Field(..., description="1-sentence description of language use.")

class UserPersona(BaseModel):
    """
    Schema representing a single user persona archetype used for simulation or evaluation.

    Fields:
    - title: Short, descriptive archetype name (e.g., “Data-Driven Analyst”, “Casual Browser”).
    - key_traits_summary: 2–3 sentence overview summarizing core traits, behaviors, and mindset.
    - background: 1–2 sentence summary providing professional or demographic context.
    - tech_savviness: Nested object describing technology familiarity and rationale.
    - primary_goal: The main objective, motivation, or job-to-be-done for this persona.
    - language_proficiency: Nested object detailing English fluency and communication characteristics.
    - behavioral_quirks: 1–2 sentence outline of notable behavioral or interaction quirks.
    """
    title: str = Field(..., description="Short, descriptive archetype.")
    key_traits_summary: str = Field(..., description="2–3 sentence summary of core traits, behaviors, mindset.")
    background: str = Field(..., description="1–2 sentence professional/demographic context.")
    tech_savviness: TechSavviness
    primary_goal: str = Field(..., description="Specific job-to-be-done or primary task.")
    language_proficiency: LanguageProficiency
    behavioral_quirks: str = Field(..., description="1–2 sentence interaction style and behavior.")

class PersonasPayload(BaseModel):
     """
    Schema for the complete persona payload containing a fixed set of persona objects.

    Fields:
    - user_personas: List containing exactly 15 persona entries.
      Each persona object defines traits, background, tech savviness, goals, language proficiency,
      and behavioral quirks, collectively forming a diverse evaluation or simulation set.
    """
     user_personas: conlist(UserPersona, min_length=15, max_length=15) = Field(
        ..., description="List of 15 persona objects."
    )

def qaresponse_to_df_personas(qa_response):
    """
    Convert a PersonasPayload object into a flat DataFrame.

    Each row corresponds to a single UserPersona entry from PersonasPayload.user_personas.

    The function:
    - Accepts either a Pydantic v1/v2 model instance or a plain dictionary.
    - Flattens nested persona fields (tech_savviness, language_proficiency) into top-level DataFrame columns.
    - Produces a tabular structure suitable for inspection, export, or analysis.

    Returns:
        pd.DataFrame: A DataFrame with one row per persona and columns for:
        Title, Key Traits Summary, Background, Tech Savviness Level, Tech Savviness Justification,
        Primary Goal, Language Proficiency Level, Language Proficiency Characteristic, and Behavioral Quirks.
    """
    # Get the list of personas from either model or dict
    if hasattr(qa_response, "user_personas"):
        personas = qa_response.user_personas
    elif isinstance(qa_response, dict):
        personas = qa_response.get("user_personas", [])
    else:
        personas = []

    def _val(obj, attr, default=None):
        # Robustly read attribute from pydantic model or dict
        if obj is None:
            return default
        if isinstance(obj, dict):
            return obj.get(attr, default)
        return getattr(obj, attr, default)

    rows = []
    for item in personas:
        tech = _val(item, "tech_savviness", {})  # nested: TechSavviness
        lang = _val(item, "language_proficiency", {})  # nested: LanguageProficiency

        # Support both dict and model objects for nested fields
        tech_level = _val(tech, "level")
        tech_just = _val(tech, "justification")
        lang_level = _val(lang, "level")
        lang_char = _val(lang, "characteristic")

        rows.append({
            "Title": _val(item, "title", ""),
            "Key Traits Summary": _val(item, "key_traits_summary", ""),
            "Background": _val(item, "background", ""),
            "Tech Savviness Level": tech_level,
            "Tech Savviness Justification": tech_just,
            "Primary Goal": _val(item, "primary_goal", ""),
            "Language Proficiency Level": lang_level,
            "Language Proficiency Characteristic": lang_char,
            "Behavioral Quirks": _val(item, "behavioral_quirks", ""),
        })

    return pd.DataFrame(rows, columns=[
        "Title",
        "Key Traits Summary",
        "Background",
        "Tech Savviness Level",
        "Tech Savviness Justification",
        "Primary Goal",
        "Language Proficiency Level",
        "Language Proficiency Characteristic",
        "Behavioral Quirks",
    ])

def get_personas(business_use_case):
    """
    Generate a structured set of user personas for a given business use case.

    Workflow:
    - Sends the provided business use case description to the model with a persona generation prompt.
    - Receives a structured PersonasPayload (validated Pydantic model) containing 15 personas.
    - Converts the structured response into a flat DataFrame for analysis or export.

    Inputs:
    - business_use_case: A text description outlining the product, target audience, or business context
      used to guide persona generation.

    Returns:
    - DataFrame of 15 personas, each with fields such as Title, Key Traits Summary, Background,
      Tech Savviness, Language Proficiency, Primary Goal, and Behavioral Quirks.
    """
    resp1 = get_openai_resp_struct(system_prompt_persona_generation, user_prompt_personas, business_use_case, PersonasPayload)
    df1 = qaresponse_to_df_personas(resp1)

    return df1



### generate questions based on each persona

def get_QA_personas(dfct, df_personas, n=5):
    """
    Generate QA items conditioned on different user personas.

    Workflow:
    - Iterates through each persona from the provided persona DataFrame.
    - Dynamically formats a persona-specific system prompt using persona attributes
      (traits, background, goals, tech savviness, and language proficiency).
    - For each persona, samples random document chunks and queries the model
      to generate persona-aligned QA items.
    - Aggregates responses across personas into a unified DataFrame with
      a 'check_metric' column identifying the generating persona.

    Inputs:
    - dfct: DataFrame containing text chunks with 'chunk_id' and 'text' fields.
    - df_personas: DataFrame of persona definitions (flattened from PersonasPayload).
    - n: Number of questions to generate per persona (function batches calls accordingly).

    Returns:
    - DataFrame of generated QA items tagged by persona, with columns:
      ['Question', 'Answer', 'Chunk IDs', 'Difficulty', 'check_metric'].
    """
    df_r = pd.DataFrame(columns=['Question', 'Answer', 'Chunk IDs', 'Difficulty', 'check_metric'])
    for persona_gen in tqdm(json.loads(df_personas.to_json(orient='records')), desc='Running different personas'):
        system_prompt_t = system_prompt_persona_specific_QA.format(
            title=persona_gen['Title'],
            key_traits_summary=persona_gen['Key Traits Summary'],
            background=persona_gen['Background'],
            tech_savviness_level=persona_gen['Tech Savviness Level'],
            tech_savviness_justification=persona_gen['Tech Savviness Justification'],
            primary_goal=persona_gen['Primary Goal'],
            language_proficiency_level=persona_gen['Language Proficiency Level'],
            language_proficiency_characteristic=persona_gen['Language Proficiency Characteristic'],
            behavioural_quirks=persona_gen['Behavioral Quirks']
        )
        for i1 in range(int(np.ceil(n/5))):
            sample_chunks = dfct.sample(20)
            cd = sample_chunks[['chunk_id', 'text']].to_dict(orient='records')
            resp1 = get_openai_resp_struct(system_prompt_t, user_prompt_chunks, json.dumps(cd), QAResponse_basic)
            df1 = qaresponse_to_df_basic(resp1)
            df1['check_metric'] = 'persona_'+str(persona_gen['Title'])
            df_r = pd.concat([df_r, df1])
            time.sleep(1)
    return df_r


def generate_and_save(data_path, usecase = None, save_path_dir='test_dataset', n_queries=100):
    """
    Orchestrate comprehensive test dataset generation from PDFs, including persona-based QA creation,
    and save all intermediate and final artifacts to disk.

    Steps performed:
    - Load PDF files from `data_path` using DirectoryLoader / PyPDFLoader.
    - Semantically chunk documents (via chunking_utilities.semantically_chunk_documents).
    - Convert chunks to a DataFrame, assign chunk IDs, and perform semantic clustering.
    - Persist the clustered dataset as 'chunked_dataset.csv'.
    - Generate multiple classes of QA items:
        (1) Basic QA items (general comprehension),
        (2) Chunk-length sensitivity QA,
        (3) Chunk-boundary synthesis QA,
        (4) Query-intent evaluation QA,
        (5) Persona-conditioned QA (if `usecase` is provided).
    - Automatically generate 15 personas (from the given business use case) and save them as
      'personas_generated.csv' before creating persona-aligned QA items.
    - Concatenate all QA outputs, standardize columns, and save the unified dataset as
      'test_data_generated.csv' under `save_path_dir`.

    Inputs:
    - data_path: Directory path containing PDF source documents.
    - usecase: Optional business use case description used to generate personas and persona-specific QA.
    - save_path_dir: Output directory where all intermediate and final CSVs will be saved.
    - n_queries: Total number of QA items to generate (distributed across all generation modes).

    Returns:
    - A success message string after successfully generating and saving all datasets.

    Notes:
    - Requires helper functions from `chunking_utilities` (semantically_chunk_documents, chunks_to_df, cluster_chunks_df).
    - The persona generation step executes only if `usecase` is provided.
    - The function dynamically balances generation volume across five QA types to achieve variety and coverage.
    """
    os.makedirs(save_path_dir, exist_ok=True)
    loader = DirectoryLoader(data_path, glob = '*.pdf', loader_cls = PyPDFLoader)
    docs = loader.load()
    
    chunks = chunking_utilities.semantically_chunk_documents(
        docs,                                  # same input as before
        model_name="sentence-transformers/all-MiniLM-L6-v2",  # small, fast model
        min_tokens=80,                         # prevent overly small chunks
        max_tokens=350,                        # keep chunks within retriever budget
        similarity_threshold=0.6,              # cohesion control; higher = stricter
        overlap_sentences=1,                   # carry 1 sentence into next chunk
    )

    dfc = chunking_utilities.chunks_to_df(chunks)

    dfc['chunk_id'] = dfc.index


    dfc1 = chunking_utilities.cluster_chunks_df(dfc)
    dfc1.to_csv(save_path_dir+'/'+'chunked_dataset.csv', index=False)
    num_t = 20 ### number of cohorst of generation = 15 personas + 4 edge cases = 19 ~= 20
    for iterationi in tqdm(range(5), desc='Generating test data'):
        if iterationi==0:
            df1 = get_QA_basic(dfc1.copy(), n=np.ceil(n_queries/num_t).astype(int))
        elif iterationi==1:
            df2 = get_QA_chunk_length(dfc1.copy(), n=np.ceil(n_queries/num_t).astype(int))
        elif iterationi==2:
            df3 = get_QA_chunk_boundary(dfc1.copy(), n=np.ceil(n_queries/num_t).astype(int))
        elif iterationi==3:
            df4 = get_QA_query_intent(dfc1.copy(), n=np.ceil(n_queries/num_t).astype(int))
        elif usecase:
            business_use_case = usecase
            df_personas = get_personas(business_use_case)
            df_personas.to_csv(save_path_dir+'/'+'personas_generated.csv', index=False)
            df5 = get_QA_personas(dfc1.copy(), df_personas.copy(), n=np.ceil(n_queries/num_t).astype(int))

    df1['Rationale'] = 'None'
    df1['check_metric'] = 'general'
    df2['check_metric'] = 'chunk_length'
    df3['check_metric'] = 'chunk_boundary'
    df4['check_metric'] = 'user_intent'
    df5['Rationale'] = 'None'
    
    df1['Less Relevant Chunk IDs'] = np.nan
    df3['Less Relevant Chunk IDs'] = np.nan
    df4['Less Relevant Chunk IDs'] = np.nan
    df5['Less Relevant Chunk IDs'] = np.nan
    
    df1 = df1[['Question', 'Answer', 'Chunk IDs', 'Less Relevant Chunk IDs', 'Difficulty', 'Rationale', 'check_metric']]
    df2 = df2[['Question', 'Answer', 'Chunk IDs', 'Less Relevant Chunk IDs', 'Difficulty', 'Rationale', 'check_metric']]
    df3 = df3[['Question', 'Answer', 'Chunk IDs', 'Less Relevant Chunk IDs', 'Difficulty', 'Rationale', 'check_metric']]
    df4 = df4[['Question', 'Answer', 'Chunk IDs', 'Less Relevant Chunk IDs', 'Difficulty', 'Rationale', 'check_metric']]
    df5 = df5[['Question', 'Answer', 'Chunk IDs', 'Less Relevant Chunk IDs', 'Difficulty', 'Rationale', 'check_metric']]

    df_test = pd.concat([df1, df2, df3, df4, df5])

    df_test = df_test.reset_index(drop=True)

    df_test.to_csv(save_path_dir+'/'+'test_data_generated.csv', index=False)
    return "Created and Saved successfully"

# generate_and_save(data_path=r'../data1/',
#                 usecase='chatbot to guide data scientists on how to test their AI Agents',
#                 save_path='test_dataset',
#                 n_queries=100)



import traceback

from vero.metrics import MetricBase
import numpy as np
import math
# from rag_code.tracing_components import logger


class MeanRR(MetricBase):
    '''
        Implementation of MRR for reranker performance measure.

        :param reranked_docs: Pass the list of reranked documents list.
        :param ground_truth: Pass the ground truth list for comparison.

        Methods
        ---------
        1. __init__(reranked_docs, ground_truth, k)
            Initializes the metric.
        2. evaluate() -> list
            Returns the mrr score.

        :returns: float
        '''
    name = 'mean_reciprocal_rank'

    def __init__(self,reranked_docs: list, ground_truth: list):
        self.reranked_docs = reranked_docs
        self.ground_truth = ground_truth

    def evaluate(self) -> float | None:
        # logger.info('Starting MRR calculation...')
        reciprocal_rank = []
        count = 0
        try:
            for docs, truth in zip(self.reranked_docs, self.ground_truth):
                count += 1
                for i in range(len(docs)):
                    if docs[i] in truth:
                        reciprocal_rank.append(1 / (i + 1))
                        break
                if len(reciprocal_rank) < count:
                    reciprocal_rank.append(0)

            mrr = round((sum(reciprocal_rank) / len(reciprocal_rank)), 2)
            return mrr
        except Exception as e:
            # logger.info('Exception occured during MRR calculation\nError:', e)
            print('Exception occured during MRR calculation\nError:', traceback.format_exc())
            return None


import traceback

from vero.metrics import MetricBase
import gc
import os
import numpy as np
from typing import Optional
from openai import OpenAI
import torch
from dotenv import load_dotenv, find_dotenv
load_dotenv(find_dotenv())
# from tracing_components import logger


# Check if a GPU is available and set the device accordingly
device = "cuda" if torch.cuda.is_available() else "cpu"

#TODO:expose openai client for model config
class GEvalScore(MetricBase):
    '''
        A unique implementation of g-eval.

        :param api_key: Pass the api key for OpenAI client.

        Methods
        ---------
        1. __init__(api_key)
            Initializes the client.

        2. evaluate() -> float
            Returns the g-eval score.
        '''
    name = 'g_eval_score'

    def __init__(self,api_key:str | None = None):
        api_key = os.getenv('OPENAI_API_KEY')
        self.api_key = api_key
        self.client = OpenAI(api_key=api_key)

    def __enter__(self):
        return self

    # tiktoken for tokens, extracting logits
    def evaluate(self,reference, candidate, metric, metric_description:str =None, rubric :str =None, custom_prompt: Optional[str] = None, polling: bool = False, polling_calls: int = 5) -> float | None:
        '''
        :param reference: Pass the reference chunks.
        :param candidate: Pass the answer.
        :param metric: Pass the metric.
        :param metric_description: Pass the metric description. (Optional)
        :param rubric: Pass the rubric. (Optional)
        :param custom_prompt: Pass the custom prompt. (Optional)
        :param polling: Pass if polling is enabled. (Optional)
        :param polling_calls: Pass the number of polling calls. (Optional)

        :return: float
        '''
        client = self.client
        if rubric is None:
            rubric = '1 to 5; 1 being worst and 5 being best'
        score = 0
        pc = polling_calls
        log_list = {}
        g_eval_score = 0
        logits = {32: 10, 33: 10, 34: 10, 35: 10, 36: 10}

        meta_prompt = f'''
                You are an expert AI prompt engineer specializing in creating evaluation prompts for Large Language Models (LLMs) based on the G-Eval framework. Your task is to generate a comprehensive and structured evaluation prompt template based on user-provided requirements.

                Instructions:
                    Role-Play: The generated prompt must begin by assigning a clear role to the evaluator LLM (e.g., "You are an expert evaluator.").

                    Task Definition: Clearly define the evaluation task using the user's Metric Name and Metric Description (description is optional).

                    Context Placeholders:

                    Always include placeholders(give curly braces so placeholders can be filled) for Retrieved Context (i.e. ref) and Generated Output (i.e. candidate).

                    Evaluation Criteria: Integrate the user's Rating Scale. Provide a detailed definition for what each level of the scale represents.

                    Reasoning Requirement: MANDATE a step-by-step reasoning process (Chain-of-Thought) to evaluate and provide the final score.

                    Structured Output: Instruct the LLM to provide its final output in a clean, easily parsable form and the output should ONLY include the scale provided by the user and nothing else.

                These are the requirements given by the user:
                Metric - {metric}
                Metric Description - {metric_description}
                Rating rubric - {rubric}      

                **Crucial** only include the prompt don't include something like:'Sure, Here is the prompt for this' or 'Here's a comprehensive and structured evaluation prompt template based on the provided requirements:' or 'Fill in the placeholders ref and candidate with the appropriate context and generated output to utilize this prompt effectively.', etc
                Also dont include final score in curly braces as it will become a placeholder and throw error when formatting.
                **Crucial** include this line in the generated prompt - 'Your final output must be ONLY this score. Do not output any reasoning, explanations, introductory text, quotation marks, asterisks, or any other characters. The entire response must be the raw score value.'
            '''

        try:
            if custom_prompt == None:
                prompt_call = client.chat.completions.create(
                    model='gpt-4o-mini',
                    messages=[{'role': 'user', 'content': meta_prompt}],
                )
                prompt = prompt_call.choices[0].message.content
                prompt = prompt.format(ref=reference, candidate=candidate)
                print(prompt)
            else:
                prompt = custom_prompt

            while True:
                completion = client.chat.completions.create(
                    model='gpt-4o-mini',
                    messages=[{'role': 'user', 'content': prompt}],
                    logprobs=True,
                    top_logprobs=10,
                    logit_bias=logits,

                )
                if polling:
                    score += int(completion.choices[0].message.content.strip())
                    if polling_calls > 1:
                        polling_calls -= 1
                        continue
                    else:
                        break
                else:
                    break

            if polling:
                g_eval_score = score / pc
                return g_eval_score
            else:
                top_two_logprobs = completion.choices[0].logprobs.content[0].top_logprobs
                for logprob in top_two_logprobs:
                    num = logprob.token.strip()
                    if num in ['1', '2', '3', '4', '5']:
                        log_score = np.round(np.exp(logprob.logprob), 4)
                        score += log_score
                        log_list[int(num)] = log_score

                if score == 0:
                    print('Error in calculating g-eval score')
                    #logger.error('Error in generating g-eval score')
                else:
                    for i, j in log_list.items():
                        g_eval_score += i * float((j / score))
                    return g_eval_score

        except Exception as e:
            # logger.error('Error in generating g-eval score\nError', e)
            print('Error in generating g-eval score\nError:', traceback.format_exc())
            return None


    def __exit__(self, exc_type, exc_value, traceback):
        del self.client
        self.client = None
        gc.collect()
        torch.cuda.empty_cache()


report_generation_prompt='''

ROLE:
You are an expert Principal Technical Writer and AI Analyst Communicator. You specialize in translating complex, structured data from multiple analytical sources into a single, cohesive, and actionable report. Your writing is clear, concise, data-driven, and professional.

GOAL:
Your sole purpose is to synthesize all the provided structured JSON analysis from the entire RAG evaluation workflow into a single, comprehensive, and well-formatted evaluation report in Markdown format. The final report should be clear, insightful, and actionable for a technical audience of AI engineers and product managers.

INPUT DESCRIPTION:
You will receive a JSON `DATA` which is a collection of JSON objects that represent the complete, tiered analysis of the RAG pipeline. This includes:

Evaluation Context (from parser agent) : The foundational data, including the pipeline configuration and all raw results.

Retriever Analysis, Reranker Analysis, Generator Analysis (retriever, reranker and generator analyst): The detailed, qualitative diagnoses for each individual component.

Holistic Diagnosis(diagnosis agent): The high-level synthesis, including the overall narrative and identified causal chains.

Recommendations(recommender agent): The actionable suggestions for pipeline improvements and evaluation gaps.

INPUTS:
DATA: {data}

CORE TASK & WRITING INSTRUCTIONS:
You are to act as a composer, not a new analyst. Your task is to populate the report template below using only the information provided in the input JSONs.

Follow the Template Strictly: The structure of your output must follow the provided Markdown report template exactly.

Cite Your Data: This is critical. Every claim or finding you write MUST be supported by citing the specific metric and score from the input data. For example, instead of "The retriever was good," you must write, "The retriever showed excellent performance, achieving a Recall@5 of 0.92."

Maintain a Professional Tone: The language should be objective and analytical.

Do Not Invent Information: You must not add any new analysis, conclusions, or recommendations that are not explicitly present in the input JSONs. Your job is to format and present the findings of the other agents.

Markdown Report Template/Example for structure reference (Your Output Structure)**Crucial** should only include this markdown as output and nothing else.
# RAG Pipeline Advanced Evaluation Report

- **Report Date:** 29-08-2025
- **Pipeline Version:** RAG Test v1.2
- **Status:** Comprehensive Diagnostic & Action Plan

---

## 1. Executive Summary

The pipeline demonstrates strong generation faithfulness but suffers from **retrieval starvation** and **reranking inefficiencies**, causing missed knowledge grounding and degraded task success. Attribution analysis shows **52% of failures are due to retrieval recall gaps**, while **30% arise from reranking misplacements**.

**Root-Cause Attribution:**

- Retrieval → 52%
- Reranking → 30%
- Generation → 18%

## 2. Prioritized Next Steps

### Block 1: Retriever Coverage Gaps

- `Issue Identified:` <br>Retriever surfaces relevant background but misses some trial-level detail.
- `Current Output:` <br>
  **Query:** _“Does IV vitamin C help cancer patients?”_ → _“Vitamin C is an essential nutrient with redox functions and is generally well tolerated in clinical trials.”_ _[Source: Doc ID 121]_
- `Suggested Fix:`<br> Integrate **hybrid retrieval (dense + BM25)** to ensure both background context and trial results are retrieved.
- `After Fix Expected Output:` <br>_“Several studies of IV vitamin C in cancer patients reported improved quality of life and reduced side effects, though anticancer efficacy was mixed.”_ _[Source: Doc ID 121, 122]_

### Block 2: Domain Context missed

- `Issue Identified:` <br>Pipeline retrieves useful definitions but doesn’t consistently elevate concise trial evidence to the top.
- `Current Output:` <br>
  **Query:** _“Did multivitamins reduce cancer risk in trials?”_ → _“Multivitamins are commonly used in the U.S. and may influence chronic disease outcomes.”_ _[Source: Doc ID 122]_
- `Suggested Fix:`<br> **Fine-tune embeddings** on domain Q–A pairs so trial findings rank above general use-case passages.
- `After Fix Expected Output:` <br>_“The SU.VI.MAX trial showed a 31% reduction in overall cancer risk in men using vitamin C, vitamin E, β-carotene, selenium, and zinc.”_ _[Source: Doc ID 122]_

### Block 3: Reranker Bias Toward Generic Content

- `Issue Identified:` <br>Reranker favors general overviews; precise quantitative findings are ranked lower.
- `Current Output:`<br>
  **Query:** _“What is the plasma concentration of IV vitamin C?”_ → _“IV vitamin C can achieve higher concentrations than oral intake.”_ _[Source: Doc ID 121]_
- `Suggested Fix:`<br> **A/B test reranker disable/enable** and fine-tune on clinical passages (PK, trial results).
- `After Fix Expected Output:`<br> _“Pharmacokinetic studies show oral vitamin C peaks below 300 μM, whereas IV administration achieves plasma levels up to 20 mM.”_ _[Source: Doc ID 121]_

## 3. Pipeline Configuration & Scope

| Component     | Model / Method                                                                    |
| :------------ | :-------------------------------------------------------------------------------- |
| **Retriever** | VectorStoreRetriever (Qdrant; semantic similarity with k=20, score_threshold=0.7) |
| **Reranker**  | CrossEncoderReranker (BAAI/bge-reranker-large)                                    |
| **Generator** | OpenAI (gpt-4-turbo)                                                              |

## 4. Holistic Diagnosis & Root-Cause Analysis

### Strengths

- Generator achieves **faithfulness = 0.98**, rare hallucination.
- Retriever provides generally sufficient **context sufficiency = 0.84**.

### Weaknesses

- **Moderate task success (0.62)** vs faithfulness (0.98) suggests relevance, not truthfulness, is limiting user outcomes.
- **Retrieval Recall gap** (0.42) starves generator of relevant content.
- **Reranker bias** towards stale but semantically similar chunks.
- **Reranker’s** moderate ranking performance (MAP 0.50, NDCG 0.57) fails to consistently surface the most relevant chunks.

### Causal Chains

- Retriever misses domain context → semantic coverage collapse → SEMScore drop (0.64).
- Reranker misordering → correct doc buried (MRR: 0.61).
- Generator stays faithful → but “faithful to wrong or incomplete context.”

## 5. Component-Level Deep Dive

### Retriever Performance

- **Recall:** 0.42 (Poor) <br>
  _Inference:_ Large portions of relevant evidence are missed, limiting downstream quality.
- **Precision:** 0.66 (Moderate)<br>
  _Inference:_ Retrieves a mix of relevant and noisy passages, reducing clarity.
- **Length Bias Index:** +0.12 → favors longer irrelevant chunks. <br>
  _Inference:_ Overweighting long passages wastes context budget.
- **Freshness Recall:** 0.31 → stale docs retrieved over updated ones. <br>
  _Inference:_ Pipeline struggles to prioritize latest information.

### Reranker Performance

- **MAP:** 0.50 (below 0.70 target) <br>
  _Inference:_ Only half of relevant passages are consistently ranked well.
- **MRR:** 0.61 <br>
  _Inference:_ First relevant result often buried below ideal rank.
- **NDCG\@10:** 0.57 <br>
  _Inference:_ Top-10 ordering misplaces important context.

### Generator Performance

- **Faithfulness:** 0.98 (excellent) <br>
  _Inference:_ Answers align closely with retrieved context, minimal hallucination.
- **SEMScore:** 0.64 (moderate semantic alignment) <br>
  _Inference:_ Outputs partially capture intended meaning but miss nuance.
- **ROUGE-L:** 0.12 (low lexical overlap) → verbose, less extractive answers.<br>
  _Inference:_ Style drifts from source phrasing, reducing extractive trust.
- **Groundedness Density:** 0.64 (evidence coverage in answers) <br>
  _Inference:_ Only \~2/3 of generated text is explicitly grounded in sources.

## 6. Actionable Recommendations

### Retriever

- Add **hybrid retrieval (dense + BM25)** to capture both semantic and keyword matches.
- **Fine-tune embeddings** on domain Q\&A pairs for better semantic alignment.
- Use **smaller/sliding-window chunks (500–700 tokens)** to preserve context continuity.
- Benchmark **alternative embedding models** for recall/precision trade-offs.

### Reranker

- **A/B test disabling** to measure real contribution.
- **Fine-tune** on domain-specific query→passage pairs (esp. time-sensitive QAs).
- Test **lighter or alternative architectures** (bi-encoders, newer cross-encoders).
- Adjust **reranker top_n/batch size** to rescore a broader candidate pool.

### Generator

- Strengthen prompts with **“use only provided context + cite sources”** instructions.
- **Instruction-tune on multi-hop/domain QAs** for richer reasoning and higher overlap.
- Add **consistency filters (e.g., AlignScore)** to block low-faithfulness outputs.

<br>

---

_All claims are directly supported by the provided metrics and agent analyses; all scores mentioned are averages._

---

SELF-REFLECTION & FIDELITY AUDIT:
Before you finalize the Markdown report, conduct this final audit of your work:
- Data Citation: "Have I gone through the entire report and ensured that every analytical claim is followed by the specific metric and score that proves it? (e.g., '...critically undermined by a faulty reranker, evidenced by an NDCG drop from 0.85 to 0.72.')"
- Completeness: "Have I included all the key findings, diagnoses, causal chains, and recommendations from my JSON inputs? Is anything important missing?"
- Template Adherence: "Does my final output strictly follow the requested Markdown template structure, including all headings and sections?"

The final report must be a perfect, verifiable representation of the structured analysis.
'''


generation_analyst_prompt = '''
ROLE:
You are a world-class expert AI Analyst specializing in evaluating the **Answer Generation** component of RAG (Retrieval-Augmented Generation) pipelines. Your analysis is data-driven, objective, and concise.

GOAL:
Your goal is to analyze the provided evaluation results for the **Answer Generation component** and produce a qualitative diagnosis in a structured JSON format. You must identify specific strengths and weaknesses, referencing the provided metrics and scores to justify your conclusions.

INPUT DESCRIPTION:
You will receive a JSON object containing a subset of the full evaluation context. It will primarily include:
1.  `metrics_definitions`: To understand what each metric measures.
2.  `generation_evaluation_results`: A structured document (with line breaks to signify new row) with metric names and their scores specifically for the **Answer Generation component**.

INPUT:
Metrics Definitions: {metrics_definitions}
Generation Evaluation Results: {generation_evaluation_results}

ANALYSIS HEURISTICS & DOMAIN KNOWLEDGE:
To perform your analysis, you MUST use the following expert heuristics. These are your rules for determining if a score is good, moderate, or poor.

[SPECIALIZED HEURISTICS FOR THE COMPONENT]
{heuristics}

TASK:
1.  Carefully review each metric and its score in the `generation_evaluation_results` input.
2.  Compare each score against the provided "Analysis Heuristics".
3.  Identify all scores that indicate a **Strength** (i.e., fall into the "Good" or "Excellent" range).
4.  Identify all scores that indicate a **Weakness** (i.e., fall into the "Moderate" or "Poor" range).
5.  Formulate your findings into lists of human-readable sentences, ensuring each sentence mentions the specific metric and score that supports the claim.
6.  Provide a single, concise `overall_diagnosis` sentence that summarizes the component's performance.

OUTPUT SPECIFICATION:
Your final output MUST be a single, valid JSON object and nothing else. It must conform to this exact structure:
{{
  "component_name": "Answer Generation component",
  "overall_diagnosis": "A one or two sentence summary of the component's performance.",
  "strengths": [
    "A descriptive sentence about a strength, referencing the metric and score. (e.g., 'Achieved an excellent Recall@5 score of 0.92, indicating it is highly effective at finding relevant documents.')"
  ],
  "weaknesses": [
    "A descriptive sentence about a weakness, referencing the metric and score. (e.g., 'The MRR@5 of 0.65 is only moderate, suggesting the most relevant document is not always ranked first.')"
  ]
}}

CONSTRAINTS:
* If there are no clear strengths, the `strengths` array should be empty.
* If there are no clear weaknesses, the `weaknesses` array should be empty.
* Every statement in the `strengths` and `weaknesses` arrays MUST be justified with a metric and its score.
* Be precise and objective. Do not add information not present in the input.

SELF-REFLECTION & JUSTIFICATION REVIEW:
Before finalizing your JSON output, you must perform this critical self-review:
- Evidence Check: "Is every single sentence in my strengths and weaknesses arrays directly and explicitly supported by a specific metric and its score from the input? (e.g., 'The AlignScore of 0.81 is only moderate...')."
- Objectivity Check: "Does my overall_diagnosis stay within the bounds of the provided 'Analysis Heuristics'? Have I avoided making subjective claims or assumptions that are not backed by the data?"
- Clarity Check: "Is my analysis clear and unambiguous? Will the next agent understand precisely which metrics are driving my conclusions?"
- **Crucial** Output Structure: "Have I included unnecessary things before or after the required output?, Have I adhered to required schema?"

Only after confirming these points should you generate the final JSON.

'''


retriever_analyst_prompt = '''
ROLE:
You are a world-class expert AI Analyst specializing in evaluating the **Retriever** component of RAG (Retrieval-Augmented Generation) pipelines. Your analysis is data-driven, objective, and concise.

GOAL:
Your goal is to analyze the provided evaluation results for the **Retriever component** and produce a qualitative diagnosis in a structured JSON format. You must identify specific strengths and weaknesses, referencing the provided metrics and scores to justify your conclusions.

INPUT DESCRIPTION:
You will receive a JSON object containing a subset of the full evaluation context. It will primarily include:
1.  `metrics_definitions`: To understand what each metric measures.
2.  `retriever_evaluation_results`: A dictionary of metric names and their scores specifically for the **Retriever component**.

INPUTS:
Metrics Definitions: {metrics_definitions}
Retriever Evaluation Results: {retriever_evaluation_results}

ANALYSIS HEURISTICS & DOMAIN KNOWLEDGE:
To perform your analysis, you MUST use the following expert heuristics. These are your rules for determining if a score is good, moderate, or poor.

[SPECIALIZED HEURISTICS FOR THE COMPONENT]
{heuristics}

TASK:
1.  Carefully review each metric and its score in the `retriever_evaluation_results` input.
2.  Compare each score against the provided "Analysis Heuristics".
3.  Identify all scores that indicate a **Strength** (i.e., fall into the "Good" or "Excellent" range).
4.  Identify all scores that indicate a **Weakness** (i.e., fall into the "Moderate" or "Poor" range).
5.  Formulate your findings into lists of human-readable sentences, ensuring each sentence mentions the specific metric and score that supports the claim.
6.  Provide a single, concise `overall_diagnosis` sentence that summarizes the component's performance.

OUTPUT SPECIFICATION:
Your final output MUST be a single, valid JSON object and nothing else. It must conform to this exact structure:
{{
  "component_name": "Retriever component",
  "overall_diagnosis": "A one-sentence summary of the component's performance.",
  "strengths": [
    "A descriptive sentence about a strength, referencing the metric and score. (e.g., 'Achieved an excellent Recall@5 score of 0.92, indicating it is highly effective at finding relevant documents.')"
  ],
  "weaknesses": [
    "A descriptive sentence about a weakness, referencing the metric and score. (e.g., 'The MRR@5 of 0.65 is only moderate, suggesting the most relevant document is not always ranked first.')"
  ]
}}

CONSTRAINTS:
* If there are no clear strengths, the `strengths` array should be empty.
* If there are no clear weaknesses, the `weaknesses` array should be empty.
* Every statement in the `strengths` and `weaknesses` arrays MUST be justified with a metric and its score.
* Be precise and objective. Do not add information not present in the input.

SELF-REFLECTION & JUSTIFICATION REVIEW:
Before finalizing your JSON output, you must perform this critical self-review:
- Evidence Check: "Is every single sentence in my strengths and weaknesses arrays directly and explicitly supported by a specific metric and its score from the input? (e.g., 'The AlignScore of 0.81 is only moderate...')."
- Objectivity Check: "Does my overall_diagnosis stay within the bounds of the provided 'Analysis Heuristics'? Have I avoided making subjective claims or assumptions that are not backed by the data?"
- Clarity Check: "Is my analysis clear and unambiguous? Will the next agent understand precisely which metrics are driving my conclusions?"
- **Crucial** Output Structure: "Have I included unnecessary things before or after the required output?, Have I adhered to required schema?"

Only after confirming these points should you generate the final JSON.
'''




reranker_analyst_prompt = '''
ROLE:
You are a world-class expert AI Analyst specializing in evaluating the **Reranker** component of RAG (Retrieval-Augmented Generation) pipelines. Your analysis is data-driven, objective, and concise.

GOAL:
Your goal is to analyze the provided evaluation results for the **Reranker component** and produce a qualitative diagnosis in a structured JSON format. You must identify specific strengths and weaknesses, referencing the provided metrics and scores to justify your conclusions.

INPUT DESCRIPTION:
You will receive a JSON object containing a subset of the full evaluation context. It will primarily include:
1.  `metrics_definitions`: To understand what each metric measures.
2.  `reranker_evaluation_results`: A dictionary of metric names and their scores specifically for the **Reranker component**.

INPUTS:
Metrics Definitions: {metrics_definitions}
Reranker Evaluation Results: {reranker_evaluation_results}

ANALYSIS HEURISTICS & DOMAIN KNOWLEDGE:
To perform your analysis, you MUST use the following expert heuristics. These are your rules for determining if a score is good, moderate, or poor.

[SPECIALIZED HEURISTICS FOR THE COMPONENT]
{heuristics}

TASK:
1.  Carefully review each metric and its score in the `reranker_evaluation_results` input.
2.  Compare each score against the provided "Analysis Heuristics".
3.  Identify all scores that indicate a **Strength** (i.e., fall into the "Good" or "Excellent" range).
4.  Identify all scores that indicate a **Weakness** (i.e., fall into the "Moderate" or "Poor" range).
5.  Formulate your findings into lists of human-readable sentences, ensuring each sentence mentions the specific metric and score that supports the claim.
6.  Provide a single, concise `overall_diagnosis` sentence that summarizes the component's performance.

OUTPUT SPECIFICATION:
Your final output MUST be a single, valid JSON object and nothing else. It must conform to this exact structure:
{{
  "component_name": "Reranker component",
  "overall_diagnosis": "A one-sentence summary of the component's performance.",
  "strengths": [
    "A descriptive sentence about a strength, referencing the metric and score. (e.g., 'Achieved an excellent Recall@5 score of 0.92, indicating it is highly effective at finding relevant documents.')"
  ],
  "weaknesses": [
    "A descriptive sentence about a weakness, referencing the metric and score. (e.g., 'The MRR@5 of 0.65 is only moderate, suggesting the most relevant document is not always ranked first.')"
  ]
}}

CONSTRAINTS:
* If there are no clear strengths, the `strengths` array should be empty.
* If there are no clear weaknesses, the `weaknesses` array should be empty.
* Every statement in the `strengths` and `weaknesses` arrays MUST be justified with a metric and its score.
* Be precise and objective. Do not add information not present in the input.

SELF-REFLECTION & JUSTIFICATION REVIEW:
Before finalizing your JSON output, you must perform this critical self-review:
- Evidence Check: "Is every single sentence in my strengths and weaknesses arrays directly and explicitly supported by a specific metric and its score from the input? (e.g., 'The AlignScore of 0.81 is only moderate...')."
- Objectivity Check: "Does my overall_diagnosis stay within the bounds of the provided 'Analysis Heuristics'? Have I avoided making subjective claims or assumptions that are not backed by the data?"
- Clarity Check: "Is my analysis clear and unambiguous? Will the next agent understand precisely which metrics are driving my conclusions?"
- **Crucial** Output Structure: "Have I included unnecessary things before or after the required output?, Have I adhered to required schema?"

Only after confirming these points should you generate the final JSON.
'''


import traceback

from vero.metrics import MetricBase
import gc
from rouge_score import rouge_scorer
import torch
# from tracing_components import logger


# Check if a GPU is available and set the device accordingly
device = "cuda" if torch.cuda.is_available() else "cpu"

class RougeScore(MetricBase):
    '''
        Calculates ROUGEL Score which helps in measuring recall (focuses on the Longest Common Subsequence (LCS)).

        Methods
        ---------
        1.  __init__()
            Initializes the model for metric.

        2. evaluate(reference,candidate) -> tuple
            Returns the rouge score tuple which contains precision, recall and f1 score.

        :returns: tuple
        '''
    name = 'rouge_score'

    def __init__(self):
        self.rg_scorer = rouge_scorer.RougeScorer(['rougeL'], use_stemmer=True)

    def __enter__(self):
        return self

    def evaluate(self,reference: str | list, candidate: str | list) -> tuple | None:
        '''
        :param reference: Pass the chunks for reference.
        :param candidate: Pass the answer that has to be checked.

        :return: tuple
        '''
        # logger.info('Starting ROUGE Score calculation')
        rg_scorer = self.rg_scorer
        try:
            if isinstance(reference, str):
                score = rg_scorer.score(reference, candidate)
                return (round(score['rougeL'].precision, 2), round(score['rougeL'].recall, 2),
                        round(score['rougeL'].fmeasure, 2))
            elif isinstance(reference, list):
                sum_p, sum_r, sum_f1 = 0, 0, 0
                if isinstance(candidate, list):
                    candidate = candidate[0]
                for doc in reference:
                    score = rg_scorer.score(doc, candidate)
                    sum_p += score['rougeL'].precision
                    sum_r += score['rougeL'].recall
                    sum_f1 += score['rougeL'].fmeasure
                avg_p = sum_p / len(reference)
                avg_f1 = sum_f1 / len(reference)
                avg_r = sum_r / len(reference)
                return (round(avg_p, 2), round(avg_r, 2), round(avg_f1, 2))

        except Exception as e:
            # logger.info('ROUGE score calculation failed\nError:', e)
            print('Error calculating ROUGE Score\nError:', traceback.format_exc())
            return None

    def __exit__(self, exc_type, exc_value, traceback):
        del self.rg_scorer
        self.rg_scorer = None
        gc.collect()
        torch.cuda.empty_cache()


from .inference import Inferencer
from typing import List

class AlignScorer:
    def __init__(self, model: str, batch_size: int, device: int, ckpt_path: str, evaluation_mode='nli_sp', verbose=True) -> None:
        self.model = Inferencer(
            ckpt_path=ckpt_path, 
            model=model,
            batch_size=batch_size, 
            device=device,
            verbose=verbose
        )
        self.model.nlg_eval_mode = evaluation_mode

    def score(self, contexts: List[str], claims: List[str]) -> List[float]:
        return self.model.nlg_eval(contexts, claims)[1].tolist()


from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast


VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt", "tokenizer_file": "tokenizer.json"}


PRETRAINED_VOCAB_FILES_MAP = {
    "vocab_file": {
        "lucadiliello/bleurt-tiny-128": "https://huggingface.co/lucadiliello/bleurt-tiny-128/resolve/main/vocab.txt",
        "lucadiliello/bleurt-tiny-512": "https://huggingface.co/lucadiliello/bleurt-tiny-512/resolve/main/vocab.txt",
        "lucadiliello/bleurt-base-128": "https://huggingface.co/lucadiliello/bleurt-base-128/resolve/main/vocab.txt",
        "lucadiliello/bleurt-base-512": "https://huggingface.co/lucadiliello/bleurt-base-512/resolve/main/vocab.txt",
        "lucadiliello/bleurt-large-128": "https://huggingface.co/lucadiliello/bleurt-large-128/resolve/main/vocab.txt",
        "lucadiliello/bleurt-large-512": "https://huggingface.co/lucadiliello/bleurt-large-512/resolve/main/vocab.txt",
    },
    "tokenizer_file": {
        "lucadiliello/bleurt-tiny-128": "https://huggingface.co/lucadiliello/bleurt-tiny-128/resolve/main/tokenizer.json",  # noqa: E501
        "lucadiliello/bleurt-tiny-512": "https://huggingface.co/lucadiliello/bleurt-tiny-512/resolve/main/tokenizer.json",  # noqa: E501
        "lucadiliello/bleurt-base-128": "https://huggingface.co/lucadiliello/bleurt-base-128/resolve/main/tokenizer.json",  # noqa: E501
        "lucadiliello/bleurt-base-512": "https://huggingface.co/lucadiliello/bleurt-base-512/resolve/main/tokenizer.json",  # noqa: E501
        "lucadiliello/bleurt-large-128": "https://huggingface.co/lucadiliello/bleurt-large-128/resolve/main/tokenizer.json",  # noqa: E501
        "lucadiliello/bleurt-large-512": "https://huggingface.co/lucadiliello/bleurt-large-512/resolve/main/tokenizer.json",  # noqa: E501
    },
}


PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    "lucadiliello/bleurt-base-128": 128,
    "lucadiliello/bleurt-base-512": 512,
    "lucadiliello/bleurt-large-128": 128,
    "lucadiliello/bleurt-large-512": 512,
    "lucadiliello/bleurt-tiny-128": 128,
    "lucadiliello/bleurt-tiny-512": 512,
}


PRETRAINED_INIT_CONFIGURATION = {
    "lucadiliello/bleurt-base-128": {"do_lower_case": True},
    "lucadiliello/bleurt-base-512": {"do_lower_case": True},
    "lucadiliello/bleurt-large-128": {"do_lower_case": True},
    "lucadiliello/bleurt-large-512": {"do_lower_case": True},
    "lucadiliello/bleurt-tiny-128": {"do_lower_case": True},
    "lucadiliello/bleurt-tiny-512": {"do_lower_case": True},
}


class BleurtTokenizerFast(BertTokenizerFast):
    r"""
    Construct a "fast" BLEURT tokenizer (backed by HuggingFace's *tokenizers* library). Based on WordPiece.
    This class is identical to `BertTokenizerFast` apart from the default values for tokenizer files,
    initial configurations and positional embeddings defaults.

    This tokenizer inherits from [`PreTrainedTokenizerFast`] which contains most of the main methods. Users should
    refer to this superclass for more information regarding those methods.

    Args:
        vocab_file (`str`):
            File containing the vocabulary.
        do_lower_case (`bool`, *optional*, defaults to `True`):
            Whether or not to lowercase the input when tokenizing.
        unk_token (`str`, *optional*, defaults to `"[UNK]"`):
            The unknown token. A token that is not in the vocabulary cannot be converted to an ID and is set to be this
            token instead.
        sep_token (`str`, *optional*, defaults to `"[SEP]"`):
            The separator token, which is used when building a sequence from multiple sequences, e.g. two sequences for
            sequence classification or for a text and a question for question answering. It is also used as the last
            token of a sequence built with special tokens.
        pad_token (`str`, *optional*, defaults to `"[PAD]"`):
            The token used for padding, for example when batching sequences of different lengths.
        cls_token (`str`, *optional*, defaults to `"[CLS]"`):
            The classifier token which is used when doing sequence classification (classification of the whole sequence
            instead of per-token classification). It is the first token of the sequence when built with special tokens.
        mask_token (`str`, *optional*, defaults to `"[MASK]"`):
            The token used for masking values. This is the token used when training this model with masked language
            modeling. This is the token which the model will try to predict.
        clean_text (`bool`, *optional*, defaults to `True`):
            Whether or not to clean the text before tokenization by removing any control characters and replacing all
            whitespaces by the classic one.
        tokenize_chinese_chars (`bool`, *optional*, defaults to `True`):
            Whether or not to tokenize Chinese characters. This should likely be deactivated for Japanese (see [this
            issue](https://github.com/huggingface/transformers/issues/328)).
        strip_accents (`bool`, *optional*):
            Whether or not to strip all accents. If this option is not specified, then it will be determined by the
            value for `lowercase` (as in the original BERT).
        wordpieces_prefix (`str`, *optional*, defaults to `"##"`):
            The prefix for subwords.
    """

    vocab_files_names = VOCAB_FILES_NAMES
    pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    pretrained_init_configuration = PRETRAINED_INIT_CONFIGURATION
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES



import traceback

from vero.metrics import MetricBase
import gc
from .bartscore import BARTScorer
import torch
# from tracing_components import logger


# Check if a GPU is available and set the device accordingly
device = "cuda" if torch.cuda.is_available() else "cpu"

class BartScore(MetricBase):
    '''
        Calculates Bart Score which helps in comparing two different models/configurations.
        Doesn't have significance on its own, it's a comparison score.

        Methods
        ---------
        1.  __init__()
            Initializes the model for metric.

        2. evaluate(reference,candidate,batch_size) -> float
            Returns the bart score.

        :returns: float
    '''
    name = 'bart_score'

    def __init__(self):
        self.bart_scorer = BARTScorer()

    def __enter__(self):
        return self

    def evaluate(self,reference: str | list, candidate: str | list, batch_size: int = 4) -> float | None:
        '''
        :param reference: Pass the chunks for reference.
        :param candidate: Pass the answer that has to be checked.
        :param batch_size: Pass the number of samples to evaluate. (Optional)

        :returns: float
        '''

        # logger.info('Starting Bart Score calculation')
        bart_scorer = self.bart_scorer
        score = 0
        try:
            if isinstance(reference, str):
                score = bart_scorer.score([reference], [candidate], batch_size=batch_size)
                return round(score[0], 2)
            elif isinstance(reference, list):
                sum_score = 0
                if isinstance(candidate, str):
                    candidate = [candidate]
                for doc in reference:
                    score = bart_scorer.score([doc], candidate, batch_size=batch_size)
                    sum_score += score[0]
                avg_score = sum_score / len(reference)
                return round(avg_score, 2)
        except Exception as e:
            print('Error calculating Bart Score\nError:', traceback.format_exc())
            # logger.error('Error calculating Bart Score\nError:', e)

            return None
        # score = scorer.multi_ref_score([ref],[candidate],batch_size = 6)


    def __exit__(self, exc_type, exc_value, traceback):
        del self.bart_scorer
        self.bart_scorer = None
        gc.collect()
        torch.cuda.empty_cache()


from vero.metrics import MetricBase
import gc
from bert_score import BERTScorer
import torch
import traceback
# from tracing_components import logger


# Check if a GPU is available and set the device accordingly
device = "cuda" if torch.cuda.is_available() else "cpu"

class BertScore(MetricBase):
    '''
        Calculates Bert Score which compares tokens of candidate and reference texts via contextual embeddings (like BERT), computing precision, recall, and F1 to assess semantic similarity beyond exact word overlap.

        Methods
        ---------
        1.  __init__()
            Initializes the model for metric.

        2. evaluate(reference,candidate) -> tuple
            Returns the bert score tuple which contains precision, recall and f1 score.

        :returns: tuple
        '''
    name = 'bert_score'

    def __init__(self):
        self.bert_scorer = BERTScorer(model_type='bert-base-uncased', device=device)

    def __enter__(self):
        return self

    def evaluate(self,reference: str | list, candidate: str | list) -> tuple | None:
        '''
        :param reference: Pass the chunks for reference.
        :param candidate: Pass the answer that has to be checked.

        :return: tuple
        '''
        # logger.info('Starting BERTscore calculation')
        bert_scorer = self.bert_scorer
        try:
            if isinstance(reference, str):
                p, r, f1 = bert_scorer.score([candidate], [reference])
                return (round(p.item(), 2), round(r.item(), 2), round(f1.item(), 2))
            elif isinstance(reference, list):
                sum_p, sum_r, sum_f1 = 0, 0, 0
                if isinstance(candidate, str):
                    candidate = [candidate]
                for doc in reference:
                    p, r, f1 = bert_scorer.score(candidate, [doc])
                    sum_p += p.item()
                    sum_r += r.item()
                    sum_f1 += f1.item()
                avg_p = sum_p / len(reference)
                avg_f1 = sum_f1 / len(reference)
                avg_r = sum_r / len(reference)
                return (round(avg_p, 2), round(avg_r, 2), round(avg_f1, 2))
        except Exception as e:
            print('Error calculating Bert Score\nError:', traceback.format_exc())
            # logger.info('BERTscore calculation failed\nError:', e)
            return None

    def __exit__(self, exc_type, exc_value, traceback):
        del self.bert_scorer
        self.bert_scorer = None
        gc.collect()
        torch.cuda.empty_cache()